# encoding: utf-8

"""Provides abstracted cube-measure objects used as the basis for second-order measures.

There are several cube-measures that can appear in a cube-response, including
unweighted-counts, weighted-counts (aka. counts), means, and others.
"""

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty


class CubeMeasures:
    """Provides access to all cube-measure objects for this cube-result."""

    def __init__(self, cube, dimensions, slice_idx):
        self._cube = cube
        self._dimensions = dimensions
        self._slice_idx = slice_idx

    @lazyproperty
    def cube_means(self):
        """_BaseCubeMeans subclass object for this cube-result."""
        return _BaseCubeMeans.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def cube_overlaps(self):
        """_BaseCubeOverlaps subclass object for this cube-result."""
        return _BaseCubeOverlaps.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def cube_sum(self):
        """_BaseCubeSums subclass object for this cube-result."""
        return _BaseCubeSums.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def cube_stddev(self):
        """_BaseCubeStdDev subclass object for this cube-result."""
        return _BaseCubeStdDev.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def unconditional_cube_counts(self):
        """_BaseUnconditionalCounts subclass object for this cube-result."""
        return _BaseUnconditionalCubeCounts.factory(
            self._cube, self._dimensions, self._slice_idx
        )

    @lazyproperty
    def unweighted_cube_counts(self):
        """_BaseWeightedCounts subclass object for this cube-result."""
        valid_counts = self._cube.unweighted_valid_counts
        counts = (
            valid_counts if valid_counts is not None else self._cube.unweighted_counts
        )
        diff_nans = True if valid_counts is not None else False

        return _BaseCubeCounts.factory(
            counts, diff_nans, self._cube, self._dimensions, self._slice_idx
        )

    @lazyproperty
    def weighted_cube_counts(self):
        """_BaseWeightedCounts subclass object for this cube-result."""
        valid_counts = self._cube.weighted_valid_counts
        counts = valid_counts if valid_counts is not None else self._cube.counts
        diff_nans = True if valid_counts is not None else False

        return _BaseCubeCounts.factory(
            counts, diff_nans, self._cube, self._dimensions, self._slice_idx
        )


class _BaseCubeMeasure:
    """Base class for all cube-measure objects."""

    def __init__(self, dimensions):
        self._dimensions = dimensions

    @classmethod
    def _slice_idx_expr(cls, cube, slice_idx):
        """Return np.s_ advanced-indexing slice object to extract values for slice_idx.

        `cube` can contain data for multiple `_Slice` objects. The returned `numpy`
        advanced-indexing expression selects only those values of `cube` that pertain
        to the slice indicated by `slice_idx`.
        """
        # --- for a 2D cube we take the whole thing (1D is not expected here) ---
        if cube.ndim < 3:
            return np.s_[:]

        # --- if 0th dimension of a >2D cube is MR, we only take the "Selected" portion
        # --- of the indicated initial-MR subvar, because the slice is to represent the
        # --- values for "respondents who selected" that MR response (and not those who
        # --- didn't select it or did not respond).
        if cube.dimension_types[0] == DT.MR:
            return np.s_[slice_idx, 0]

        # --- for other 3D cubes we just select the 2D "table" portion associated with
        # --- the `slice_idx`-th table dimension element.
        return np.s_[slice_idx]


# === COUNTS (UNWEIGHTED & WEIGHTED) ===


class _BaseCubeCounts(_BaseCubeMeasure):
    """Base class for count (weighted & unweighted) cube-measure variants."""

    def __init__(self, dimensions, counts, diff_nans):
        super(_BaseCubeCounts, self).__init__(dimensions)
        self._counts = counts
        self._diff_nans = diff_nans

    @classmethod
    def factory(cls, counts, diff_nans, cube, dimensions, slice_idx):
        """Return _BaseCubeCounts subclass instance appropriate to `cube`

        Chooses between unweighted and weighted counts based on `type`.
        """
        dimension_type_strings = tuple(
            "MR"
            if dim_type == DT.MR
            else "ARR"
            if dim_type in DT.ARRAY_TYPES
            else "CAT"
            for dim_type in cube.dimension_types[-2:]
        )
        CubeCountsCls = {
            ("MR", "MR"): _MrXMrCubeCounts,
            ("MR", "ARR"): _MrXArrCubeCounts,
            ("MR", "CAT"): _MrXCatCubeCounts,
            ("ARR", "MR"): _ArrXMrCubeCounts,
            ("ARR", "ARR"): _ArrXArrCubeCounts,
            ("ARR", "CAT"): _ArrXCatCubeCounts,
            ("CAT", "MR"): _CatXMrCubeCounts,
            ("CAT", "ARR"): _CatXArrCubeCounts,
        }.get(dimension_type_strings, _CatXCatCubeCounts)

        return CubeCountsCls(
            dimensions, counts[cls._slice_idx_expr(cube, slice_idx)], diff_nans
        )

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.column_bases`"
        )

    @lazyproperty
    def columns_base(self):
        """Optional 1D np.float64 ndarray of column-wise base for each column.

        This marginal is used as the (un)weighted N row that is shown (typically)
        below other measures.

        This value is None if the columns dimension is not CAT because in that case each
        row has a unique rows base, and so it is not possible to reduce to a 1D
        vector in the columns orientation (aka a single row). Subclasses with legitimite
        values (eg *_X_CAT) must override this default method.
        """
        return None

    @lazyproperty
    def columns_pruning_mask(self):
        """1D bool np.ndarray indicating whether all cells in column are empty

        The columns-pruning-mask indicates when every cell in a column has base values
        of 0. When the column is *not* an MR, the base used is equal to the columns_base,
        but when the column is an MR, we use the sum of selected & non-selected values from
        the column, whereas the columns_base uses just the selected values.
        """
        # --- Bases are positive, so we can sum them to see if all of them are equal to 0
        return self._columns_pruning_base == 0

    @lazyproperty
    def columns_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each column.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, but each component is meaningful:

        * "columns": Indicates it is a marginal in the "columns" orientation (kind of
          like a stripe in the shape of a row).
        * "table": Indicates that it is the base for the whole table. When the
          `.table_base` exists (CAT X CAT), it is a repetition of that, but when
          the rows are array (and therefore we can't sum across them), each cell has
          its own value.
        * "base": Indicates that it is the base, not necessarily the counts (eg the sum
          of selected and non-selected for MR variables)

        This value is None if the columns dimension is not CAT because in that case each
        row has a unique table base, and so it is not possible to reduce to a 1D
        vector in the columns orientation (aka a single row). Subclasses with legitimite
        values (eg *_X_CAT) must override this default method.
        """
        return None

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.counts`"
        )

    @lazyproperty
    def diff_nans(self):
        """Boolean, indicates if subtotal differences (rows and cols) has to be NaN.

        We set the subtotal differences to be NaN for count measures that are paired
        alongside numeric dimension collapsing measures (eg mean/count/etc.) because
        those measures are undefined for subtotal differences. We want to do this here
        because we want to make sure that all calculations derived from the counts are
        also correctly set to NaN, rather than having to track down all the derived
        calculations.

        The parameter must be passed in during _BaseCubeCount's creation because it has
        no knowledge of what type of counts it's using, and the way to tell if the counts
        are paired with a numeric variable is by knowing if it comes from `*_counts`
        or `*_valid_counts`.
        """
        return self._diff_nans

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.row_bases`"
        )

    @lazyproperty
    def rows_base(self):
        """Optional 1D np.float64 ndarray of row-wise base for each row.

        This marginal is used as the (un)weighted N column that is shown alongside other
        measures.

        This value is None if the rows dimension is not CAT because in that case each
        column has a unique rows base, and so it is not possible to reduce to a 1D
        vector in the rows orientation (aka a single column). Subclasses with legitimite
        values (eg CAT_X_*) must override this default method.
        """
        return None

    @lazyproperty
    def rows_pruning_mask(self):
        """1D bool np.ndarray indicating whether all cells in rows are empty

        The rows-pruning-mask indicates when every cell in a row has base values
        of 0. When the row is *not* an MR, the base used is equal to the rows_base,
        but when the row is an MR, we use the sum of selected & non-selected values from
        the row, whereas the rows_base uses just the selected values.
        """
        # --- Bases are positive, so we can sum them to see if all of them are equal
        # --- to 0.
        return self._rows_pruning_base == 0

    @lazyproperty
    def rows_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each row.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, but each component is meaningful:

        * "rows": Indicates it is a marginal in the "rows" orientation (kind of
          like a stripe in the shape of a row).
        * "table": Indicates that it is the base for the whole table. When the
          `.table_base` exists (CAT X CAT), it is a repetition of that, but when
          the columns are array (and therefore we can't sum across them), each cell has
          its own value.
        * "base": Indicates that it is the base, not necessarily the counts (eg the sum
          of selected and non-selected for MR variables)

        This value is None if the rows dimension is not CAT because in that case each
        column has a unique table base, and so it is not possible to reduce to a 1D
        vector in the rows orientation (aka a single column). Subclasses with legitimite
        values (eg CAT_X_*) must override this default method.
        """
        return None

    @lazyproperty
    def table_base(self):
        """Optional float of the table-wise base for the whole table

        None if either dimension is an ARRAY or MR because the table base varies.
        The subclass with a legitimite value (CAT_X_CAT) must override this default
        method.
        """
        return None

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.table_bases`"
        )

    @lazyproperty
    def _columns_pruning_base(self):
        """1D bool np.ndarray of the sum of the column-bases for cells in a column

        Used to calculate the columns-pruning-mask, which indicates when every cell in a
        column has base values of 0. When the column is *not* an MR, the base used is
        equal to the columns_base, but when the column is an MR, we use the sum of
        selected & non-selected values the column, whereas the columns_base uses just the
        selected values. These values are not meaningful on their own, but used to
        calculate the columns-pruning-mask.

        This method works when the column is not an MR, and must be overriden when it is
        MR.
        """
        # --- Bases are positive, so we can sum them to see if all of them are equal to 0
        # --- To really hammer home the point that this number is invalid, we keep it
        # --- private. But it's easier to test, so it is separate from the mask.
        return np.sum(self.column_bases, axis=0)

    @lazyproperty
    def _rows_pruning_base(self):
        """1D bool np.ndarray of the sum of the row-bases for cells in a row

        Used to calculate he rows-pruning-mask, which indicates when every cell in a row
        of 0. When the row is *not* an MR, the base used is equal to the rows_base,
        but when the row is an MR, we use the sum of selected & non-selected values from
        the row, whereas the rows_base uses just the selected values. These values
        are not meaningful on their own, but used to calculate the rows-pruning-mask.

        This method works when the row is not an MR, and must be overriden when it is MR.
        """
        # --- Bases are positive, so we can sum them to see if all of them are equal to 0
        # --- To really hammer home the point that this number is invalid, we keep it
        # --- private. But it's easier to test, so it is separate from the mask.
        return np.sum(self.row_bases, axis=1)


class _ArrXArrCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=ARR & columns=ARR dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, bases are equal to counts
        return self.counts

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- No MR, so counts are already in correct shape
        return self._counts

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, bases are equal to counts
        return self.counts

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, bases are equal to counts
        return self.counts


class _ArrXCatCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=ARR & columns=CAT dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, bases are equal to counts
        return self.counts

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- No MR, so counts are already in correct shape
        return self._counts

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        return np.broadcast_to(self.rows_base[:, None], self.counts.shape)

    @lazyproperty
    def rows_base(self):
        """Optional 1D np.float64 ndarray of row-wise base for each row.

        This marginal is used as the (un)weighted N column that is shown alongside other
        measures.
        """
        # --- Avaialable because column is CAT, equal to the sum across cols dimension.
        return np.sum(self._counts, axis=1)

    @lazyproperty
    def rows_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each row.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, see `_BaseCubeCounts.rows_table_base` for details.
        """
        # --- Available because columns are CAT, equal to the rows_base because the row
        # --- is array and so addition over rows is not possible
        return self.rows_base

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- Can't sum across rows array, table bases equal to rows bases
        return self.row_bases


class _ArrXMrCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=ARR & columns=MR dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        return self.counts

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- Only selected from the selection dimension
        return self._counts[:, :, 0]

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        # --- selected and not-selected both contribute to margin (axis=2), both rows
        # --- and columns are retained.
        return np.sum(self._counts, axis=2)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, but use row bases to add
        # --- over the selected/not selected dimension
        return self.row_bases

    @lazyproperty
    def _columns_pruning_base(self):
        """1D bool np.ndarray of the sum of the row-bases for cells in a row.

        Used to compute the columns-pruning-mask; not meaningful on its own.
        """
        # --- Because column is MR we need to override. We want to sum over the
        # --- selection dimension (dim=2) & the rows (dim=0)
        return np.sum(self._counts, axis=(0, 2))


class _CatXArrCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=CAT & columns=ARR dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        return np.broadcast_to(self.columns_base, self.counts.shape)

    @lazyproperty
    def columns_base(self):
        """Optional 1D np.float64 ndarray of column-wise base for each column.

        This marginal is used as the (un)weighted N row that is shown (typically)
        below other measures.
        """
        # --- Avaialable because row is CAT, equal to the sum across rows dimension.
        return np.sum(self._counts, axis=0)

    @lazyproperty
    def columns_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each column.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, see `_BaseCubeCounts.columns_table_base` for details.
        """
        # --- Available because row is CAT, equal to columns_base because column is ARR
        return self.columns_base

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- No MR, so counts are already in correct shape
        return self._counts

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, bases are equal to counts
        return self.counts

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- Can't sum across columns array, table bases equal to column bases
        return self.column_bases


class _CatXCatCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=CAT & columns=CAT dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        return np.broadcast_to(self.columns_base, self.counts.shape)

    @lazyproperty
    def columns_base(self):
        """Optional 1D np.float64 ndarray of column-wise base for each column.

        This marginal is used as the (un)weighted N row that is shown (typically)
        below other measures.
        """
        # --- Avaialable because row is CAT, equal to the sum across rows dimension.
        return np.sum(self._counts, axis=0)

    @lazyproperty
    def columns_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each column.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, see `_BaseCubeCounts.columns_table_base` for details.
        """
        # --- Available because rows are CAT, equal to a repeat of the scalar table base
        # --- because both dimensions are CAT.
        return np.repeat(self.table_base, self.counts.shape[1])

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- No MR, so counts are already in correct shape
        return self._counts

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        return np.broadcast_to(self.rows_base[:, None], self.counts.shape)

    @lazyproperty
    def rows_base(self):
        """Optional 1D np.float64 ndarray of row-wise base for each row.

        This marginal is used as the (un)weighted N column that is shown alongside other
        measures.
        """
        # --- Avaialable because column is CAT, equal to the sum across cols dimension.
        return np.sum(self._counts, axis=1)

    @lazyproperty
    def rows_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each row.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, see `_BaseCubeCounts.rows_table_base` for details.
        """
        # --- Available because columns are CAT, equal to a repeat of the scalar table
        # --- base because both dimensions are CAT.
        return np.repeat(self.table_base, self.counts.shape[0])

    @lazyproperty
    def table_base(self):
        """Optional float of the table-wise base for the whole table."""
        # --- Available because both dimensions are CAT, equal to sum of counts
        return np.sum(self._counts)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- Scalar table_base, broadcast it to the correct shape
        return np.broadcast_to(self.table_base, self.counts.shape)


class _CatXMrCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=CAT & columns=MR dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        return np.broadcast_to(self.columns_base, self.counts.shape)

    @lazyproperty
    def columns_base(self):
        """Optional 1D np.float64 ndarray of column-wise base for each column.

        This marginal is used as the (un)weighted N row that is shown (typically)
        below other measures.
        """
        # --- Avaialable because row is CAT, equal to the sum across rows dimension.
        return np.sum(self.counts, axis=0)

    @lazyproperty
    def columns_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each column.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, see `_BaseCubeCounts.columns_table_base` for details.
        """
        return np.sum(self._counts, axis=(0, 2))

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- No MR, so counts are already in correct shape
        return self._counts[:, :, 0]

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        # --- selected and not-selected both contribute to margin (axis=2), both rows
        # --- and columns are retained.
        return np.sum(self._counts, axis=2)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- weighted-counts is (rows, cols, selected/not) so axis 1 is preserved to
        # --- provide a distinct value for each MR subvar.
        return np.broadcast_to(self.columns_table_base, self.counts.shape)

    @lazyproperty
    def _columns_pruning_base(self):
        """1D bool np.ndarray of the sum of the row-bases for cells in a row.

        Used to compute the columns-pruning-mask; not meaningful on its own.
        """
        # --- Because column is MR we need to override. We want to sum over the
        # --- selection dimension (dim=2) & the rows (dim=0)
        return np.sum(self._counts, axis=(0, 2))


class _MrXArrCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=MR & columns=ARR dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        # --- selected and not-selected both contribute to margin (axis=2), both rows
        # --- and columns are retained.
        return np.sum(self._counts, axis=1)

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- Only selected from the selection dimension
        return self._counts[:, 0, :]

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        return self.counts

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- No addition across subvariables possible, but use column bases to add
        # --- over the selected/not selected dimension
        return self.column_bases

    @lazyproperty
    def _rows_pruning_base(self):
        """1D bool np.ndarray of the sum of the column-bases for cells in a column.

        Used to compute the rows-pruning-mask; not meaningful on its own.
        """
        # --- Because row is MR we need to override. We want to sum over the
        # --- selection dimension (dim=1) & the columns (dim=1)
        return np.sum(self._counts, axis=(1, 2))


class _MrXCatCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=MR & columns=CAT dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        # --- selected and not-selected both contribute to margin (axis=1), both rows
        # --- and columns are retained.
        return np.sum(self._counts, axis=1)

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each valid matrix cell."""
        # --- Only selected from the selection dimension
        return self._counts[:, 0, :]

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        return np.broadcast_to(self.rows_base[:, None], self.counts.shape)

    @lazyproperty
    def rows_base(self):
        """Optional 1D np.float64 ndarray of row-wise base for each row.

        This marginal is used as the (un)weighted N column that is shown alongside other
        measures.
        """
        # --- Avaialable because column is CAT, equal to the sum across cols dimension.
        return np.sum(self.counts, axis=1)

    @lazyproperty
    def rows_table_base(self):
        """Optional 1D np.float64 ndarray of table-wise base for each row.

        This marginal is used as the denominator of the margin-table-proportions. The
        name is a mouthful, see `_BaseCubeCounts.rows_table_base` for details.
        """
        # --- Since the rows-dimension is MR, each row has a distinct table base, since
        # --- not all of the multiple responses were necessarily offered to all
        # --- respondents. The table-base for each row indicates the weighted number
        # --- of respondents who were offered that option.
        return np.sum(self._counts, axis=(1, 2))

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- Broadcast the rows_table_base to the correct shape
        return np.broadcast_to(self.rows_table_base[:, None], self.counts.shape)

    @lazyproperty
    def _rows_pruning_base(self):
        """1D bool np.ndarray of the sum of the column-bases for cells in a column.

        Used to compute the rows-pruning-mask; not meaningful on its own.
        """
        # --- Because row is MR we need to override. We want to sum over the
        # --- selection dimension (dim=1) & the columns (dim=1)
        return np.sum(self._counts, axis=(1, 2))


class _MrXMrCubeCounts(_BaseCubeCounts):
    """Counts cube-measure for a slice with rows=MR & columns=MR dimensions"""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise bases for each matrix cell."""
        # --- only column-selected counts contribute ([:, :, :, 0]), row-selected and
        # --- not-selected are summed (axis=1), rows and columns are retained.
        return np.sum(self._counts[:, :, :, 0], axis=1)

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of the count for each matrix cell."""
        # --- Only selected from the selection dimension
        return self._counts[:, 0, :, 0]

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-wise bases for each matrix cell."""
        # --- only selecteds in rows contribute ([:, 0, :, :]), selected and not from
        # --- columns both contribute (axis=2 after rows sel/not axis is collapsed)
        return np.sum(self._counts[:, 0, :, :], axis=2)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-wise bases for each matrix cell."""
        # --- Reduce second and fourth axes (the two MR_CAT dimensions) with sum()
        # --- producing 2D (nrows, ncols). This sums the (selected, selected),
        # --- (selected, not), (not, selected) and (not, not) cells of the subtable for
        # --- each matrix cell.
        return np.sum(self._counts, axis=(1, 3))

    @lazyproperty
    def _columns_pruning_base(self):
        """1D bool np.ndarray of the sum of the row-bases for cells in a row.

        Used to compute the columns-pruning-mask; not meaningful on its own.
        """
        return np.sum(self._counts[:, :, :, 0], axis=(0, 1))

    @lazyproperty
    def _rows_pruning_base(self):
        """1D bool np.ndarray of the sum of the column-bases for cells in a column.

        Used to compute the rows-pruning-mask; not meaningful on its own.
        """
        return np.sum(self._counts[:, 0, :, :], axis=(1, 2))


# === MEANS ===


class _BaseCubeMeans(_BaseCubeMeasure):
    """Base class for mean cube-measure variants."""

    def __init__(self, dimensions, means):
        super(_BaseCubeMeans, self).__init__(dimensions)
        self._means = means

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeMeans subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-means measure.
        """
        if cube.means is None:
            raise ValueError("cube-result does not contain cube-means measure")
        dimension_types = cube.dimension_types[-2:]
        CubeMeansCls = (
            _MrXMrCubeMeans
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatCubeMeans
            if dimension_types[0] == DT.MR
            else _CatXMrCubeMeans
            if dimension_types[1] == DT.MR
            else _CatXCatCubeMeans
        )
        return CubeMeansCls(
            dimensions, cube.means[cls._slice_idx_expr(cube, slice_idx)]
        )

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of cube means."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.means`"
        )


class _CatXCatCubeMeans(_BaseCubeMeans):
    """Means cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        return self._means


class _CatXMrCubeMeans(_BaseCubeMeans):
    """Means cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimensions need not actually be CAT.
    """

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        return self._means[:, :, 0]


class _MrXCatCubeMeans(_BaseCubeMeans):
    """Means cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT.
    """

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        return self._means[:, 0, :]


class _MrXMrCubeMeans(_BaseCubeMeans):
    """Means cube-measure for an MR_X_MR slice."""

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._means[:, 0, :, 0]


# === OVERLAPS ===


class _BaseCubeOverlaps(_BaseCubeMeasure):
    """Base class for overlap cube-measure variants."""

    def __init__(self, dimensions, overlaps, valid_overlaps):
        super(_BaseCubeOverlaps, self).__init__(dimensions)
        self._overlaps = overlaps
        self._valid_overlaps = valid_overlaps

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeOverlaps subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-overlaps measure
        or if it doesn't include valid-cube-overlaps measure.
        """
        if cube.overlaps is None:
            raise ValueError(
                "cube-result does not contain cube-overlaps measure"
            )  # pragma: no cover
        if cube.valid_overlaps is None:
            raise ValueError(
                "cube-result does not contain cube-valid-overlaps measure"
            )  # pragma: no cover

        dimension_types = tuple(d.dimension_type for d in dimensions)
        idx_expr = cls._slice_idx_expr(cube, slice_idx)
        args = (dimensions, cube.overlaps[idx_expr], cube.valid_overlaps[idx_expr])
        return (
            _MrXMrOverlaps(*args)
            if dimension_types == (DT.MR, DT.MR)
            else _CatXMrOverlaps(*args)
        )

    @lazyproperty
    def overlaps(self):
        """3D np.float64 ndarray of cube overlaps."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.overlaps`"
        )


class _CatXMrOverlaps(_BaseCubeOverlaps):
    """Overlaps cube-measure for a NOT_MR_X_MR slice."""

    @lazyproperty
    def tile_repetitions(self):
        """tuple of number of repetitions of selected base matrix."""
        return (self._overlaps.shape[0], 1, 1)

    @lazyproperty
    def selected_bases(self):
        """3D np.float64 ndarray of selected overlaps between MR subvariables, per cat.

        For a CAT x MR matrix, the overlaps are calculated for each category, and then
        for each subvariables pair for that category (which will produce a square
        matrix for each category). So the output shape that we get back from the
        database is CAT x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one being the result
        of the `cube_overlap` measure, and representing the "pairing" with each subvar
        of the previous MR_SUBVAR dimension.

        From this shape, we only need the "Selected" part of the MR_SEL dimension, so
        we need to select the 0th element along the 2nd axis [:, :, 0]. But we also need
        to add all of the categories together, since that's what's used for the base
        value in all of the significance calculations. We then tile these bases, since
        they're the same for each row (all the categories added together). The tiling is
        done so that the API of this class can be uniform for all users.
        """
        return np.tile(np.sum(self._overlaps[:, :, 0], axis=0), self.tile_repetitions)

    @lazyproperty
    def valid_bases(self):
        """3D np.float64 ndarray of valid overlaps between MR subvariables, per cat.

        For a CAT x MR matrix, the overlaps are calculated for each category, and then
        for each subvariables pair for that category (which will produce a square
        matrix for each category). So the output shape that we get back from the
        database is CAT x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one being the result
        of the `cube_overlap` measure, and representing the "pairing" with each subvar
        of the previous MR_SUBVAR dimension.

        From this shape, we only need the "Selected" + "Other" part of the MR_SEL
        dimension (i.e. all except missing), so we need to add the 0th and the 1st
        element along the 2nd axis sum([:, :, 0:2]).
        """
        valids = np.sum(self._valid_overlaps[:, :, 0:2], axis=2)
        return np.tile(np.sum(valids, axis=0), self.tile_repetitions)


class _MrXMrOverlaps(_BaseCubeOverlaps):
    """Overlaps cube-measure for a MR_X_MR slice."""

    @lazyproperty
    def selected_bases(self):
        """3D np.float64 ndarray of selected overlaps bases between MR subvariables.

        For a MR x MR slice, the overlaps are calculated for each MR subvar row, and
        then for each subvariables pair for that row (which will produce a square
        matrix for each category). So the output shape that we get back from the
        database is MR_SUBVAR x MR_SEL x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one
        being the result of the `cube_overlap` measure, and representing the "pairing"
        with each subvar of the last MR_SUBVAR dimension.

        From this shape, we only need the "Selected" part of the last MR_SEL dimension,
        but we need both "Selected" and "Other" from the first MR_SEL dimension. We
        therefore need to select the 0th element along the 3rd axis, but we need to sum
        the 0th and 1st elements along the 1st axis (first MR_SEL dimension). The sum
        of the selected and other counts of the first MR_SEL is what's always used to
        represent bases of such crosstabs (we cannot simply ignore the other counts).
        """
        return np.sum(self._overlaps[:, 0:2, :, 0], axis=1)

    @lazyproperty
    def valid_bases(self):
        """3D np.float64 ndarray of valid overlaps between MR subvariables, per row.

        For a MR x MR slice, the overlaps are calculated for each row, and then
        for each subvariables pair for that row (which will produce a square matrix for
        each row). So the output shape that we get back from the database is
        MR_SUBVAR x MR_SEL x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one being the
        result of the `cube_overlap` measure, and representing the "pairing" with each
        subvar of the last MR_SUBVAR dimension.

        From this shape, we only need the "Selected" + "Other" part of both MR_SEL
        dimensions (i.e. all except missing), so we need to add the 0th and the 1st
        element along the 1st and 3rd axes sum([:, 0:2, :, 0:2], axis=(1, 3)).
        """
        return np.sum(self._valid_overlaps[:, 0:2, :, 0:2], axis=(1, 3))


# === STD DEV ===


class _BaseCubeStdDev(_BaseCubeMeasure):
    """Base class for stddev cube-measure variants."""

    def __init__(self, dimensions, stddev):
        super(_BaseCubeStdDev, self).__init__(dimensions)
        self._stddev = stddev

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeStdDev subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-stddev measure.
        """
        if cube.stddev is None:
            raise ValueError("cube-result does not contain cube-stddev measure")
        dimension_types = cube.dimension_types[-2:]
        CubeSumsCls = (
            _MrXMrCubeStdDev
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatCubeStdDev
            if dimension_types[0] == DT.MR
            else _CatXMrCubeStdDev
            if dimension_types[1] == DT.MR
            else _CatXCatCubeStdDev
        )
        return CubeSumsCls(
            dimensions, cube.stddev[cls._slice_idx_expr(cube, slice_idx)]
        )

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of cube stddev."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.stddev`"
        )


class _CatXCatCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        return self._stddev


class _CatXMrCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimensions need not actually be CAT.
    """

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        return self._stddev[:, :, 0]


class _MrXCatCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT.
    """

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        return self._stddev[:, 0, :]


class _MrXMrCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for an MR_X_MR slice."""

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._stddev[:, 0, :, 0]


# === SUMS ===


class _BaseCubeSums(_BaseCubeMeasure):
    """Base class for sum cube-measure variants."""

    def __init__(self, dimensions, sums):
        super(_BaseCubeSums, self).__init__(dimensions)
        self._sums = sums

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeSums subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-sum measure.
        """
        if cube.sums is None:
            raise ValueError("cube-result does not contain cube-sum measure")
        dimension_types = cube.dimension_types[-2:]
        CubeSumsCls = (
            _MrXMrCubeSums
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatCubeSums
            if dimension_types[0] == DT.MR
            else _CatXMrCubeSums
            if dimension_types[1] == DT.MR
            else _CatXCatCubeSums
        )
        return CubeSumsCls(dimensions, cube.sums[cls._slice_idx_expr(cube, slice_idx)])

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of cube sum."""
        raise NotImplementedError(  # pragma: no cover
            f"`{type(self).__name__}` must implement `.sum`"
        )


class _CatXCatCubeSums(_BaseCubeSums):
    """Sums cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        return self._sums


class _CatXMrCubeSums(_BaseCubeSums):
    """Sums cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimensions need not actually be CAT.
    """

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        return self._sums[:, :, 0]


class _MrXCatCubeSums(_BaseCubeSums):
    """Sums cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT.
    """

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        return self._sums[:, 0, :]


class _MrXMrCubeSums(_BaseCubeSums):
    """Sums cube-measure for an MR_X_MR slice."""

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._sums[:, 0, :, 0]


# === (WEIGHTED) UNCONDITIONAL COUNTS ===


class _BaseUnconditionalCubeCounts(_BaseCubeMeasure):
    """Base class for unconditional-(weighted)-count cube-measure variants.

    Unconditional counts are the weighted counts including the missing (invalid) rows/
    columns. This cube measure is used to calculate the column-index. The name does
    not include "weighted" because we do not (currently) ever need the unweighted
    version of the unconditional counts, but they could in theory be calculated.
    """

    def __init__(
        self,
        dimensions,
        counts_with_missings,
    ):
        super(_BaseUnconditionalCubeCounts, self).__init__(dimensions)
        self._counts_with_missings = counts_with_missings

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseWeightedCounts subclass instance appropriate to `cube`."""
        dimension_types = cube.dimension_types[-2:]
        counts_with_missings = cube.counts_with_missings

        # --- TODO: This probably needs Array type subclasses like the other counts
        # --- cube measures have, because we shouldn't be adding across subvariables
        # --- I don't think we ever actually use this, though, and would also need to
        # --- choose between weighted_counts and valid_weighted_counts
        UnconditionalCubeCountsCls = (
            _MrXMrUnconditionalCubeCounts
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatUnconditionalCubeCounts
            if dimension_types[0] == DT.MR
            else _CatXMrUnconditionalCubeCounts
            if dimension_types[1] == DT.MR
            else _CatXCatUnconditionalCubeCounts
        )
        return UnconditionalCubeCountsCls(
            dimensions,
            counts_with_missings[cls._slice_idx_expr(cube, slice_idx)],
        )

    @lazyproperty
    def _valid_row_idxs(self):
        """ndarray-style index for only valid (non-missing) rows.

        Suitable for indexing a raw measure array to include only valid rows.
        """
        return np.ix_(self._dimensions[-2].valid_elements.element_idxs)


class _CatXCatUnconditionalCubeCounts(_BaseUnconditionalCubeCounts):
    """Unconditional-counts cube-measure for a slice with no MR dimensions.

    Unconditional counts are the counts including the missing (invalid) rows/
    columns. This cube measure is used to calculate the column-index.
    """

    @lazyproperty
    def baseline(self):
        """2D np.float64 ndarray of baseline value for each row in matrix.

        The shape of the return value is (nrows, 1). The baseline value for a row is the
        proportion of all values that appear in that row. A baseline for a 4 x 3 matrix
        looks like this:

            [[0.2006734 ]
             [0.72592593]
             [0.05521886]
             [0.01818182]]

        Note that the baseline values sum to 1.0. This is because each represents the
        portion of all responses that fall in that row. This baseline value is the
        denominator of the `._column_index` computation.

        Baseline is a straightforward function of the *unconditional row margin*.
        Unconditional here means that both valid and invalid responses (to the
        columns-var question) are included. This ensures that the baseline is not
        distorted by a large number of missing responses to the columns-question.
        """
        # --- uncond_row_margin is a 1D ndarray of the weighted total observation count
        # --- involving each valid row. Counts consider both valid and invalid columns,
        # --- but are only produced for valid rows.
        uncond_row_margin = np.sum(self._counts_with_missings, axis=1)[
            self._valid_row_idxs
        ]
        return uncond_row_margin[:, None] / np.sum(uncond_row_margin)


class _CatXMrUnconditionalCubeCounts(_BaseUnconditionalCubeCounts):
    """Unconditional-counts cube-measure for a NOT_MR_X_MR slice.

    Unconditional counts are the counts including the missing (invalid) rows/
    columns. This cube measure is used to calculate the column-index.
    """

    @lazyproperty
    def baseline(self):
        """2D np.float64 (or NaN) ndarray of baseline value for each matrix cell.

        Its shape is (nrows, ncols) which corresponds to CAT_X_MR_SUBVAR.

        The baseline value is compared with the column-proportion value for each cell to
        form the column-index value. The baseline is a function of the unconditional row
        margin, which is the sum of counts across both valid and missing columns.

        For CAT_X_MR, `uncond_row_margin` sums across the MR_CAT (selected, not,
        missing) dimension to include missing values (an MR_SUBVAR element is never
        "missing": true).
        """
        # --- counts_with_missings.shape is (nall_rows, ncols, selected/not/missing).
        # --- axes[1] corresponds to the MR_SUBVAR dimension, in which there are never
        # --- "missing" subvars (so nall_cols always equals ncols for that dimension
        # --- type). uncond_row_margin selects only valid rows, retains all columns and
        # --- reduces the selected/not/missing axis by summing those three counts. Its
        # --- shape is (nrows, ncols).
        uncond_row_margin = np.sum(self._counts_with_missings, axis=2)[
            self._valid_row_idxs
        ]
        # --- uncond_table_margin sums across rows, producing 1D array of size ncols,
        # --- (although all its values are always the same).
        uncond_table_margin = np.sum(uncond_row_margin, axis=0)
        # --- division produces a 2D matrix of shape (nrows, ncols) ---
        return uncond_row_margin / uncond_table_margin


class _MrXCatUnconditionalCubeCounts(_BaseUnconditionalCubeCounts):
    """Unconditional-counts cube-measure for an MR_X_NOT_MR slice.

    Unconditional counts are the counts including the missing (invalid) rows/
    columns. This cube measure is used to calculate the column-index.
    """

    @lazyproperty
    def baseline(self):
        """2D np.float64 ndarray of baseline value for each row in matrix.

        `._baseline` is the denominator of the column-index and represents the
        proportion of the overall row-count present in each row. A cell with
        a column-proportion exactly equal to this basline will have a column-index of
        100.

        The shape of the return value is (nrows, 1). A baseline for a 4 x 3 matrix looks
        something like this:

            [[0.17935204]
             [0.33454989]
             [0.50762388]
             [0.80331259]
             [0.7996507 ]]

        Baseline is a function of the *unconditional row margin*. Unconditional here
        means that both valid and invalid responses (to the columns-var question) are
        included. This ensures that the baseline is not distorted by a large number of
        missing responses to the columns-question.
        """
        # --- unconditional row-margin is a 1D ndarray of size nrows computed by:
        # --- 1. summing across all columns: np.sum(self._counts_with_missings, axis=2)
        # --- 2. taking only selected counts: [:, 0]
        # --- 3. taking only valid rows: [self._valid_row_idxs]
        uncond_row_margin = np.sum(self._counts_with_missings, axis=2)[:, 0][
            self._valid_row_idxs
        ]
        # --- The "total" (uncond_row_table_margin) is a 1D ndarray of size nrows. Each
        # --- sum includes only valid rows (MR_SUBVAR, axis 0), selected and unselected
        # --- but not missing counts ([0:2]) of the MR_CAT axis (axis 1), and all column
        # --- counts, both valid and missing (axis 2). The rows axis (0) is preserved
        # --- because each MR subvar has a distinct table margin.
        uncond_row_table_margin = np.sum(
            self._counts_with_missings[self._valid_row_idxs][:, 0:2], axis=(1, 2)
        )
        # --- inflate shape to (nrows, 1) for later calculation convenience ---
        return (uncond_row_margin / uncond_row_table_margin)[:, None]


class _MrXMrUnconditionalCubeCounts(_BaseUnconditionalCubeCounts):
    """Weighted-counts cube-measure for an MR_X_MR slice.

    Unconditional counts are the counts including the missing (invalid) rows/
    columns. This cube measure is used to calculate the column-index.
    """

    @lazyproperty
    def baseline(self):
        """2D np.float64 ndarray of baseline value for each matrix cell.

        The shape is (nrows, ncols) and all values in a given row are the same. So
        really there are only nrows distinct baseline values, but the returned shape
        makes calculating column-index in a general way more convenient.
        """
        # --- `counts_with_missings` for MR_X_MR is 4D of size (nrows, 3, ncols, 3)
        # --- (MR_SUBVAR, MR_CAT, MR_SUBVAR, MR_CAT). Unconditional row margin:
        # --- * Takes all rows and all cols (axes 0 & 2), because MR_SUBVAR dimension
        # ---   can contain only valid elements (no such thing as "missing": true
        # ---   subvar).
        # --- * Sums selected + unselected + missing categories in second MR_CAT
        # ---   dimension (columns MR, axes[3]). Including missings here fulfills
        # ---   "unconditional" characteristic of margin.
        # --- * Takes only those totals associated with selected categories of first
        # ---    MR_CAT dimension (rows MR). ("counts" for MR are "selected" counts).
        # --- Produces a 2D (nrows, ncols) array.
        uncond_row_margin = np.sum(self._counts_with_missings[:, 0:2], axis=3)[:, 0]
        # --- Unconditional table margin is also 2D (nrows, ncols) but the values for
        # --- all columns in a row have the same value; basically each row has
        # --- a distinct table margin.
        uncond_table_margin = np.sum(self._counts_with_missings[:, 0:2], axis=(1, 3))
        # --- baseline is produced by dividing uncond_row_margin by uncond_table_margin.
        return uncond_row_margin / uncond_table_margin



# -*- coding: utf-8 -*-
#
# Crunch Cube documentation build configuration file, created by
# sphinx-quickstart on Thu Oct 19 16:17:11 2017.
#
# This file is execfile() with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime
import os
import re


def ascii_bytes_from(path, *paths):
    """
    Return the ASCII characters in the file specified by *path* and *paths*.
    The file path is determined by concatenating *path* and any members of
    *paths* with a directory separator in between.
    """
    file_path = os.path.join(path, *paths)
    with open(file_path) as f:
        ascii_bytes = f.read()
    return ascii_bytes


# -- General configuration ------------------------------------------------

# Sphinx Extensions
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx_rtd_theme",
    "sphinx.ext.todo",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
]

# Templates path
templates_path = ["_templates"]

# The suffix(es) of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"Crunch Cube"
copyright = u"2021, Crunch.io"
author = u"Crunch.io"

# Get the currenct version of cr.cube
thisdir = os.path.dirname(__file__)
init_py = ascii_bytes_from(thisdir, "..", "..", "src", "cr", "cube", "__init__.py")
version = re.search('__version__ = "([^"]+)"', init_py).group(1)

# The full version, including alpha/beta/rc tags.
release = version


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"
html_logo = "crunch_logo.png"
html_domain_indices = True

html_context = {
    "display_github": True,
    "github_user": "Crunch-io",
    "github_repo": "crunch-cube",
    "github_version": "master/docs/",
    "author": "crunch-cube Authors",
    "date": datetime.date.today().strftime("%d/%m/%y"),
}
html_theme_options = {
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    # Toc options
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "CrunchCubedoc"

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "CrunchCube.tex", u"Crunch Cube Documentation", u"Crunch.io", "manual")
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "crunchcube", u"Crunch Cube Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "CrunchCube",
        u"Crunch Cube Documentation",
        author,
        "CrunchCube",
        "One line description of project.",
        "Miscellaneous",
    )
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"https://docs.python.org/": None}



# encoding: utf-8

"""Second-order measure collection and the individual measures it composes."""

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.smoothing import Smoother
from cr.cube.stripe.cubemeasure import CubeMeasures
from cr.cube.stripe.insertion import (
    NanSubtotals,
    NegativeTermSubtotals,
    PositiveTermSubtotals,
    SumSubtotals,
)
from cr.cube.util import lazyproperty


# === MEASURE COLLECTION ===


class StripeMeasures:
    """Intended to be a singleton for a given cube-result.

    It will give the same values if duplicated, just sacrificing some time and memory
    performance. Provides access to the variety of possible second-order measure objects
    for its cube-result. All construction and computation are lazy so only actually
    requested measures consume resources.
    """

    def __init__(self, cube, rows_dimension, ca_as_0th, slice_idx):
        self._cube = cube
        self._rows_dimension = rows_dimension
        self._ca_as_0th = ca_as_0th
        self._slice_idx = slice_idx

    @lazyproperty
    def means(self):
        """_Means measure object for this stripe."""
        return _Means(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def population_proportions(self):
        """_PopulationPrortion measure object for this stripe."""
        return _PopulationProportions(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def population_proportion_stderrs(self):
        """_PopulationProportionStderrs measure object for this stripe."""
        return _PopulationProportionStderrs(
            self._rows_dimension, self, self._cube_measures
        )

    @lazyproperty
    def pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each stripe row."""
        return self._cube_measures.unweighted_cube_counts.pruning_base

    @lazyproperty
    def scaled_counts(self):
        """_ScaledCounts measure object for this stripe.

        Provides access to table-totals related to numeric-values/scaled-counts.
        """
        return _ScaledCounts(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def share_sum(self):
        """_ShareSum measure object for this stripe."""
        return _ShareSum(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def smoothed_means(self):
        """_MeansSmoothed measure object for this stripe.

        If smoothing is defined in the dimension transform a _MeansSmoothed object will
        be returned, otherwise fallback to unsmoothed measure object.
        """
        return _MeansSmoothed(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def stddev(self):
        """_StdDev measure object for this stripe."""
        return _StdDev(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def sums(self):
        """_Sums measure object for this stripe."""
        return _Sums(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def table_proportion_stddevs(self):
        """_TableProportionStddevs measure object for this stripe."""
        return _TableProportionStddevs(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def table_proportion_stderrs(self):
        """_TableProportionStderrs measure object for this stripe."""
        return _TableProportionStderrs(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def table_proportion_variances(self):
        """_TableProportionVariances measure object for this stripe."""
        return _TableProportionVariances(
            self._rows_dimension, self, self._cube_measures
        )

    @lazyproperty
    def table_proportions(self):
        """_TableProportions measure object for this stripe."""
        return _TableProportions(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def unweighted_bases(self):
        """_UnweightedBases measure object for this stripe."""
        return _UnweightedBases(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def unweighted_counts(self):
        """_UnweightedCounts measure object for this stripe."""
        return _UnweightedCounts(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def weighted_bases(self):
        """_WeightedBases measure object for this stripe."""
        return _WeightedBases(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def weighted_counts(self):
        """_WeightedCounts measure object for this stripe."""
        return _WeightedCounts(self._rows_dimension, self, self._cube_measures)

    @lazyproperty
    def _cube_measures(self):
        """CubeMeasures collection object for this cube-result.

        This collection provides access to all cube-measure objects for the cube-result.
        The collection is provided to each measure object so it can access the cube
        measures it is based on.
        """
        return CubeMeasures(
            self._cube, self._rows_dimension, self._ca_as_0th, self._slice_idx
        )


# === INDIVIDUAL MEASURES ===


class _BaseSecondOrderMeasure:
    """Base class for all second-order measure objects."""

    def __init__(self, rows_dimension, measures, cube_measures):
        self._rows_dimension = rows_dimension
        self._measures = measures
        self._cube_measures = cube_measures

    @lazyproperty
    def base_values(self):
        """1D ndarray of measure base-value for each row.

        The base values are those that correspond 1-to-1 with the cube-result values.
        The values appear in payload order.
        """
        raise NotImplementedError(
            f"`{type(self).__name__}` must implement `.base_values`"
        )  # pragma: no cover

    @lazyproperty
    def blocks(self):
        """(base_values, subtotal_values) pair comprising the "blocks" of this measure.

        Use of this default implementation assumes implementation of a `._base_values`
        and `._subtotal_values` property in the subclass. A measure which is computed
        differently can override this `.blocks` property instead of implementing those
        two components.
        """
        return (self.base_values, self.subtotal_values)

    @lazyproperty
    def subtotal_values(self):
        """1D ndarray of subtotal value for each row-subtotal."""
        raise NotImplementedError(
            f"`{type(self).__name__}` must implement `.subtotal_values`"
        )  # pragma: no cover

    @lazyproperty
    def _unweighted_cube_counts(self):
        """_BaseCubeCounts for unweighted counts subclass instance for this measure.

        Provides cube measures associated with unweighted counts, including
        unweighted-counts and bases.
        """
        return self._cube_measures.unweighted_cube_counts

    @lazyproperty
    def _weighted_cube_counts(self):
        """_BaseCubeCounts for weighted counts subclass instance for this measure.

        Provides cube measures associated with weighted counts, including
        weighted-counts and table-margin.
        """
        return self._cube_measures.weighted_cube_counts


class _SmoothedMeasure(_BaseSecondOrderMeasure):
    """Mixin providing `._smoother` property for smoothed measures."""

    @lazyproperty
    def _smoother(self):
        """BaseSmoother subtype object providing smoothing as specified in spec."""
        return Smoother.factory(self._rows_dimension)


class _Means(_BaseSecondOrderMeasure):
    """Provides the means measure for a stripe.

    Relies on the presence of a means cube-measure in the cube-result.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of mean for each row."""
        return self._cube_measures.cube_means.means

    @lazyproperty
    def subtotal_values(self):
        """1D ndarray of np.nan for each row-subtotal.

        Mean values cannot be subtotaled and each subtotal value is unconditionally
        np.nan.
        """
        return NanSubtotals.subtotal_values(self.base_values, self._rows_dimension)


class _MeansSmoothed(_Means, _SmoothedMeasure):
    """Provides the smoothed means measure for a stripe.

    Relies on the presence of a means cube-measure in the cube-result.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of smoothed mean for each row."""
        return self._smoother.smooth(self._cube_measures.cube_means.means)


class _PopulationProportions(_BaseSecondOrderMeasure):
    """Provides the population-proportions measure for a stripe.

    When the variable is a categorical-date, we do not divide the population between the
    categories, so the proportion is always 1. Otherwise it is equal to the table
    proportions.

    This is used to calculate the population_counts.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of population proportion for each row."""
        proportions = self._measures.table_proportions.base_values
        return (
            np.repeat(1, proportions.shape)
            if self._rows_dimension.dimension_type == DT.CAT_DATE
            else proportions
        )

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of population proportion for each row-subtotal."""
        proportions = self._measures.table_proportions.subtotal_values
        return (
            np.repeat(1, proportions.shape)
            if self._rows_dimension.dimension_type == DT.CAT_DATE
            else proportions
        )


class _PopulationProportionStderrs(_BaseSecondOrderMeasure):
    """Provides the population-proportion-standard-errors measure for a stripe.

    When the variable is a categorical-date, we do not divide the population between the
    categories, so the standard-error is always 0. Otherwise it is equal to the table
    proportions.

    This is used to calculate the population_counts_moe.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of population proportion for each row."""
        stderrs = self._measures.table_proportion_stderrs.base_values
        return (
            np.repeat(0, stderrs.shape)
            if self._rows_dimension.dimension_type == DT.CAT_DATE
            else stderrs
        )

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of population proportion for each row-subtotal."""
        stderrs = self._measures.table_proportion_stderrs.subtotal_values
        return (
            np.repeat(0, stderrs.shape)
            if self._rows_dimension.dimension_type == DT.CAT_DATE
            else stderrs
        )


class _ScaledCounts(_BaseSecondOrderMeasure):
    """Provides access to table-totals related to numeric-values/scaled-counts."""

    @lazyproperty
    def scale_mean(self):
        """Optional float mean of scaled row-counts.

        This value is `None` when no row-elements have numeric-value. The numeric value
        (aka. "scaled-count") for a row is its count multiplied by the numeric-value of
        its element. For example, if 100 women responded "Very Likely" and the
        numeric-value of the "Very Likely" response (element) was 4, then the
        scaled-count for that row would be 400. The scale mean is the average of those
        scale values over the total count of responses. The count of a row lacking a
        numeric value does not contribute to either the numerator or denominator of that
        computation.
        """
        # --- value is None when no row-element has been assigned a numeric value ---
        if self._numeric_values.size == 0:
            return None

        # --- Also when the total count is zero, unlikely but possible and would lead to
        # --- a divide-by-zero error in this case.
        if self._total_weighted_count == 0:
            return None

        return self._total_scaled_count / self._total_weighted_count

    @lazyproperty
    def scale_median(self):
        """Optional np.float64 median of scaled weighted-counts.

        This value is `None` when no rows have a numeric-value assigned. Responses
        without a numeric value are not considered.
        """
        # --- value is None when no row-element has been assigned a numeric value ---
        if self._numeric_values.size == 0:
            return None

        # --- convert float weighted-counts to int. I'm not sure why this needs to
        # --- account for NaN values, if you know, explain it here.
        weighted_counts = np.nan_to_num(self._weighted_counts).astype("int64")

        # --- create an array with a numeric value for each individual count (response)
        # --- that can be fed directly into np.median()
        expanded_numeric_values = np.repeat(self._numeric_values, weighted_counts)

        return np.median(expanded_numeric_values)

    @lazyproperty
    def scale_stddev(self):
        """Optional np.float64 standard-deviation of scaled weighted-counts.

        This value is `None` when no row-elements have a numeric-value. The value is in
        units of scaled-counts and indicates the dispersion of the scaled-count
        distribution from its mean (scale-mean).
        """
        # --- value is None when no row-element has been assigned a numeric value
        # --- or when scale variance is not available
        if self._numeric_values.size == 0 or self._scale_variance is None:
            return None

        return np.sqrt(self._scale_variance)

    @lazyproperty
    def scale_stderr(self):
        """Optional np.float64 standard-error of scaled weighted counts.

        This value is `None` when no rows have a numeric-value assigned.
        """
        # --- value is None when no row-element has been assigned a numeric value
        # --- or when scale variance is not available
        if self._numeric_values.size == 0 or self._scale_variance is None:
            return None

        return np.sqrt(self._scale_variance / self._total_weighted_count)

    @lazyproperty
    def _has_numeric_value(self):
        """1D bool ndarray (mask) of True for each row with a defined numeric-value."""
        return ~np.isnan(np.array(self._rows_dimension.numeric_values))

    @lazyproperty
    def _numeric_values(self):
        """1D ndarray of numeric-value for each element in rows dimension that has one.

        The items in the array can be float or int. Row elements that have not been
        assigned a numeric value are skipped. Otherwise, the values appear in payload
        order.
        """
        return np.array(self._rows_dimension.numeric_values)[self._has_numeric_value]

    @lazyproperty
    def _scale_variance(self):
        """Optional np.float64 variance of scaled weighted-counts."""
        if self.scale_mean is None:
            # --- Scale mean can be None when _total_weighted_count is 0. Therefore the
            # --- scale variance has to be None to indicate that this measure is not
            # --- available.
            return None
        return (
            np.sum(
                self._weighted_counts * pow((self._numeric_values - self.scale_mean), 2)
            )
            / self._total_weighted_count
        )

    @lazyproperty
    def _total_scaled_count(self):
        """float/int total of scaled-count for rows with a numeric-value.

        The scaled-count for a row is its weighted-count multiplied by its numeric
        value (scaling-factor).
        """
        return np.sum(self._weighted_counts * self._numeric_values)

    @lazyproperty
    def _total_weighted_count(self):
        """float/int total of weighted counts for rows with a numeric-value."""
        return np.sum(self._weighted_counts)

    @lazyproperty
    def _weighted_counts(self):
        """1D float/int ndarray of weighted_count for elements with a numeric-value.

        Counts for rows that have not been assigned a numeric value are skipped.
        Otherwise, the values appear in payload order.
        """
        return self._weighted_cube_counts.counts[self._has_numeric_value]


class _ShareSum(_BaseSecondOrderMeasure):
    """Provides the share of sum measure for a stripe.

    Relies on the presence of a sum cube-measure in the cube-result.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of share of sum for each row."""
        sums = self._cube_measures.cube_sum.sums
        return sums / np.sum(sums)

    @lazyproperty
    def subtotal_values(self):
        """1D ndarray of share of sum subtotals for each row-subtotal."""
        return SumSubtotals.subtotal_values(self.base_values, self._rows_dimension)


class _StdDev(_BaseSecondOrderMeasure):
    """Provides the stddev measure for a stripe.

    Relies on the presence of a stddev cube-measure in the cube-result.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of stddev for each row."""
        return self._cube_measures.cube_stddev.stddev

    @lazyproperty
    def subtotal_values(self):
        """1D ndarray of np.nan for each row-subtotal.

        StdDev values cannot be subtotaled and each subtotal value is unconditionally
        np.nan.
        """
        return NanSubtotals.subtotal_values(self.base_values, self._rows_dimension)


class _Sums(_BaseSecondOrderMeasure):
    """Provides the sum measure for a stripe.

    Relies on the presence of a sum cube-measure in the cube-result.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of sum for each row."""
        return self._cube_measures.cube_sum.sums

    @lazyproperty
    def subtotal_values(self):
        """1D ndarray of sum subtotals for each row-subtotal."""
        return SumSubtotals.subtotal_values(self.base_values, self._rows_dimension)


class _TableProportionStddevs(_BaseSecondOrderMeasure):
    """Provides the table-proportion standard-deviation measure for a stripe."""

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of table-prop stddev for each base row of stripe."""
        return np.sqrt(self._measures.table_proportion_variances.base_values)

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of table-prop stddev for each row subtotal."""
        return np.sqrt(self._measures.table_proportion_variances.subtotal_values)


class _TableProportionStderrs(_BaseSecondOrderMeasure):
    """Provides the table-proportion standard-error measure for a stripe."""

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of table-prop stderr for each base row of stripe."""
        variances = self._measures.table_proportion_variances.base_values
        weighted_bases = self._measures.weighted_bases.base_values
        return np.sqrt(variances / weighted_bases)

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of table-prop stderr for each row subtotal."""
        variances = self._measures.table_proportion_variances.subtotal_values
        weighted_bases = self._measures.weighted_bases.subtotal_values
        return np.sqrt(variances / weighted_bases)


class _TableProportionVariances(_BaseSecondOrderMeasure):
    """Provides the table-proportion-variances measure for a stripe."""

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of table-prop variance for each base row of stripe."""
        p = self._measures.table_proportions.base_values
        return p * (1 - p)

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of table-prop variance for each row subtotal.

        The formula for the variance is more complicated when there are negative
        terms (eg for subtotal differences). For regular subtotals, this formula
        reduces to the p * (1 - p) formula and so is equivalent.

        (1 - p)^2 * (Np / Nt) + (0 - p)^2 * (Ni / Nt) + (-1 - p)^2 * (Nn / Nt)

        Where:
        p = row/col/total-proportions (eg (Np - Nn) / Nt))
        Np = weighted-count of positive terms
        Ni = weighted-count of "ignored" (eg not negative nor positive) terms
        Nn = weighted-count of negative terms
        Nt = weighted-base (sum of all counts eg Np + Ni + Nn)
        """
        p = self._measures.table_proportions.subtotal_values

        Np = PositiveTermSubtotals.subtotal_values(
            self._measures.weighted_counts.base_values, self._rows_dimension
        )
        Nn = NegativeTermSubtotals.subtotal_values(
            self._measures.weighted_counts.base_values, self._rows_dimension
        )
        Nt = self._measures.weighted_bases.subtotal_values
        Ni = Nt - Np - Nn

        p_term = (1 - p) ** 2 * (Np / Nt)
        i_term = (0 - p) ** 2 * (Ni / Nt)
        n_term = (-1 - p) ** 2 * (Nn / Nt)

        return p_term + i_term + n_term


class _TableProportions(_BaseSecondOrderMeasure):
    """Provides the table-proportions measure for a stripe.

    Table-proportions is a 1D np.float64 ndarray of the proportion each row's weighted
    count contributes to the weighted-N of the table.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of table-proportion for each row of stripe."""
        weighted_counts = self._measures.weighted_counts.base_values
        weighted_bases = self._weighted_cube_counts.bases

        # --- note that table-margin can be either scalar or 1D ndarray. When it is an
        # --- array (stripe is MR), its shape is the same as the weighted_counts, so the
        # --- division works either way.

        # --- do not propagate divide-by-zero warnings to stderr ---
        with np.errstate(divide="ignore", invalid="ignore"):
            return weighted_counts / weighted_bases

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of sum for each row-subtotal."""
        subtotal_values = self._measures.weighted_counts.subtotal_values
        weighted_table_base = self._weighted_cube_counts.table_base

        # --- table-base is defined when stripe is MR and the division below only
        # --- An Array stripe can have no subtotals, so the right answer is always
        # --- always np.array([]) in this case.
        if weighted_table_base is None:
            return np.array([])

        # --- do not propagate divide-by-zero warnings to stderr ---
        with np.errstate(divide="ignore", invalid="ignore"):
            return subtotal_values / weighted_table_base


class _UnweightedBases(_BaseSecondOrderMeasure):
    """Provides the unweighted-bases measure for a stripe.

    unweighted-bases is a 1D np.float64 ndarray of the unweighted table-proportion
    denominator (base) for each row. This object also provides the table-base totals.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of unweighted table-proportion denominator per cell."""
        return self._unweighted_cube_counts.bases

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of subtotal value for each row-subtotal."""
        # --- Background:
        # --- 1. The base is only defined for a CAT stripe.
        # --- 2. An MR stripe can have no subtotals.
        # --- The strategy here is to broadcast the table-base to the size of the
        # --- subtotals array for CAT, and return an empty array for MR.

        # --- This initial subtotal-values array has the wrong values (unless it's
        # --- empty), but has the right shape and type.
        subtotal_values = SumSubtotals.subtotal_values(
            self.base_values, self._rows_dimension
        )
        # --- in the "no-subtotals" case, return that value, since it is both the right
        # --- value and the right dtype. Note this takes care of the MR stripe case.
        if subtotal_values.shape == (0,):
            return subtotal_values

        return np.broadcast_to(
            self._unweighted_cube_counts.table_base, subtotal_values.shape
        )

    @lazyproperty
    def table_base_range(self):
        """[min, max] np.float64 ndarray range of (total) unweighted-N for this stripe.

        A non-MR stripe will have a single base, represented by min and max being the
        same value. Any subtotals in a non-MR stripe have that same scalar table base,
        so they do not affect the (single-valued) range.

        Each row of an MR stripe has a distinct base, which is reduced to a range in
        that case. An MR stripe can have no subtotals, so those don't come into it in
        the MR case.
        """
        bases = self._unweighted_cube_counts.bases
        return np.array([np.min(bases), np.max(bases)])


class _UnweightedCounts(_BaseSecondOrderMeasure):
    """Provides the unweighted-counts measure for a stripe."""

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of unweighted-count for each stripe base-row."""
        return self._unweighted_cube_counts.counts

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of sum for each row-subtotal."""
        # --- counts don't sum on an MR dimension, but an MR stripe can have no
        # --- subtotals. This just returns an empty array in that case and we don't need
        # --- to special-case MR.
        return SumSubtotals.subtotal_values(self.base_values, self._rows_dimension)


class _WeightedBases(_BaseSecondOrderMeasure):
    """Provides the weighted-bases measure for a stripe.

    weighted-bases is a 1D np.float64 ndarray of the (weighted) table-proportion
    denominator (base) for each row.
    """

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of weighted table-proportion denominator per cell."""
        return self._weighted_cube_counts.bases

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of sum for each row-subtotal."""
        # --- Background:
        # --- 1. The base is only defined for a CAT stripe.
        # --- 2. Only a CAT stripe can have subtotals; an MR stripe can't.
        # --- The strategy here is to broadcast the table-margin to the size of the
        # --- subtotals array for CAT, and return an empty array for MR.

        # --- This initial subtotal-values array has the wrong values (unless it's
        # --- empty), but has the right shape and type.
        subtotal_values = SumSubtotals.subtotal_values(
            self.base_values, self._rows_dimension
        )
        # --- in the "no-subtotals" case, return that value, since it is both the right
        # --- value and the right dtype. Note this takes care of the MR stripe case.
        if subtotal_values.shape == (0,):
            return subtotal_values

        return np.broadcast_to(
            self._weighted_cube_counts.table_base, subtotal_values.shape
        )

    @lazyproperty
    def table_margin_range(self):
        """[min, max] np.float64 ndarray range of (total) unweighted-N for this stripe.

        A non-MR stripe will have a single margin, represented by min and max being the
        same value.

        Each row of an MR stripe has a distinct margin, which is reduced to a range in
        that case. An MR stripe can have no subtotals, so those don't enter into the
        computation in the MR case.
        """
        bases = self._weighted_cube_counts.bases
        return np.array([np.min(bases), np.max(bases)])


class _WeightedCounts(_BaseSecondOrderMeasure):
    """Provides the weighted-counts measure for a stripe."""

    @lazyproperty
    def base_values(self):
        """1D np.float64 ndarray of weighted-count for each row."""
        return self._weighted_cube_counts.counts

    @lazyproperty
    def subtotal_values(self):
        """1D np.float64 ndarray of sum for each row-subtotal."""
        # --- counts don't sum on an MR dimension, but an MR stripe can have no
        # --- subtotals. This just returns an empty array in that case and we don't need
        # --- to special-case MR.
        return SumSubtotals.subtotal_values(self.base_values, self._rows_dimension)



# encoding: utf-8

"""Unit test suite for `cr.cube.collator` module."""

import sys

import numpy as np
import pytest

from cr.cube.collator import (
    _BaseAnchoredCollator,
    _BaseCollator,
    ExplicitOrderCollator,
    PayloadOrderCollator,
    SortByValueCollator,
)
from cr.cube.dimension import Dimension, _Element, _OrderSpec, _Subtotal, _ValidElements

from ..unitutil import (
    ANY,
    call,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class Describe_BaseCollator:
    """Unit-test suite for `cr.cube.collator._BaseCollator` object."""

    def it_provides_access_to_the_dimension_element_ids_to_help(self, dimension_):
        dimension_.element_ids = (42, 24, 1, 6)
        assert _BaseCollator(dimension_, None)._element_ids == (42, 24, 1, 6)

    @pytest.mark.parametrize(
        "empty_idxs, prune, hidden_idxs, expected_value",
        (
            ((), False, (), ()),
            ((), False, (4, 5, 6), (4, 5, 6)),
            ((1, 2, 3), False, (4, 5, 6), (4, 5, 6)),
            ((), True, (), ()),
            ((), True, (4, 5, 6), (4, 5, 6)),
            ((1, 2, 3), True, (4, 5, 6), (1, 2, 3, 4, 5, 6)),
        ),
    )
    def it_knows_the_hidden_idxs_to_help(
        self, dimension_, empty_idxs, prune, hidden_idxs, expected_value
    ):
        dimension_.hidden_idxs = hidden_idxs
        dimension_.prune = prune
        collator = _BaseCollator(dimension_, empty_idxs)

        assert collator._hidden_idxs == frozenset(expected_value)

    def it_provides_access_to_the_order_spec_to_help(self, request, dimension_):
        order_spec_ = instance_mock(request, _OrderSpec)
        dimension_.order_spec = order_spec_
        assert _BaseCollator(dimension_, None)._order_spec is order_spec_

    def it_provides_access_to_the_dimension_subtotals_to_help(self, dimension_):
        assert _BaseCollator(dimension_, None)._subtotals is dimension_.subtotals

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_BaseAnchoredCollator:
    """Unit-test suite for `cr.cube.collator._BaseAnchoredCollator` object."""

    def it_provides_an_interface_classmethod(self, request, dimension_):
        _init_ = initializer_mock(request, _BaseAnchoredCollator)
        _display_order_ = property_mock(
            request, _BaseAnchoredCollator, "_display_order"
        )
        _display_order_.return_value = (-3, 0, 1, -2, 2, 3, -1)

        display_order = _BaseAnchoredCollator.display_order(dimension_, empty_idxs=(2,))

        _init_.assert_called_once_with(ANY, dimension_, (2,))
        assert display_order == (-3, 0, 1, -2, 2, 3, -1)

    @pytest.mark.parametrize(
        "base_order, insertion_order, derived_order, hidden_idxs, expected_value",
        (
            # --- base-values but no insertions nor derived ---
            (((0, 0, 0), (1, 0, 1), (2, 0, 2)), (), (), (), (0, 1, 2)),
            # --- insertions but no base-values nor derived (not expected) ---
            (
                (),
                ((sys.maxsize, 0, -3), (0, 1, -2), (sys.maxsize, 0, -1)),
                (),
                (1,),
                (-2, -3, -1),
            ),
            # --- both base-values and insertions but no derived ---
            (
                ((0, 0, 0), (1, 0, 1), (2, 0, 2), (3, 0, 3)),
                ((-1, 0, -4), (1, 1, -3), (2, -1, -2), (sys.maxsize, 0, -1)),
                (),
                (1, 3),
                (-4, 0, -3, -2, 2, -1),
            ),
            # --- base-values and derived elements (but no insertions) ---
            (
                ((0, 0, 0), (2, 0, 2), (3, 0, 3)),
                (),
                ((-1, 0, 4), (2, 1, 1)),
                (),
                (4, 0, 2, 1, 3),
            ),
        ),
    )
    def it_computes_the_display_order_to_help(
        self,
        request,
        base_order,
        insertion_order,
        derived_order,
        hidden_idxs,
        expected_value,
    ):
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_hidden_idxs",
            return_value=hidden_idxs,
        )
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_insertion_orderings",
            return_value=insertion_order,
        )
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_base_element_orderings",
            return_value=base_order,
        )
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_derived_element_orderings",
            return_value=derived_order,
        )

        assert _BaseAnchoredCollator(None, None)._display_order == expected_value

    @pytest.mark.parametrize(
        "element_order_descriptors, expected_value",
        (
            # --- no elements in dimension (not an expected case) ---
            ((), ()),
            # --- 1 element ---
            (((0, 0, 6),), ((0, 0, 0),)),
            # --- 3 elements in dimension ---
            (((0, 2, 18), (1, 0, 12), (2, 1, 5)), ((0, 0, 2), (1, 0, 0), (2, 0, 1))),
        ),
    )
    def it_computes_the_base_element_orderings_to_help(
        self, request, element_order_descriptors, expected_value
    ):
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_element_order_descriptors",
            return_value=element_order_descriptors,
        )
        collator = _BaseAnchoredCollator(None, None)

        assert collator._base_element_orderings == expected_value

    def it_has_empty_derived_element_orderings(self):
        collator = _BaseAnchoredCollator(None, None)
        assert collator._derived_element_orderings == tuple()

    def it_raises_on_element_order_descriptors_access(self):
        """Error message identifies the non-implementing subclass."""
        with pytest.raises(NotImplementedError) as e:
            _BaseAnchoredCollator(None, None)._element_order_descriptors
        assert str(e.value) == (
            "`_BaseAnchoredCollator` must implement `._element_order_descriptors`"
        )

    @pytest.mark.parametrize(
        "element_order_descriptors, expected_value",
        (
            # --- no elements in dimension (not an expected case) ---
            ((), {}),
            # --- 1 element ---
            (((0, 0, 6),), {6: 0}),
            # --- 3 elements in dimension ---
            (((0, 2, 18), (1, 0, 12), (2, 1, 5)), {18: 0, 12: 1, 5: 2}),
        ),
    )
    def it_maps_the_element_ids_to_their_position_to_help(
        self, request, element_order_descriptors, expected_value
    ):
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_element_order_descriptors",
            return_value=element_order_descriptors,
        )
        collator = _BaseAnchoredCollator(None, None)

        assert collator._element_positions_by_id == expected_value

    @pytest.mark.parametrize(
        "insertion_positions, expected_value",
        (
            # --- no inserted-subtotals in dimension ---
            ((), ()),
            # --- 1 insertion ---
            (((6, 1),), ((6, 1, -1),)),
            # --- 3 insertions ---
            (
                ((7, 1), (sys.maxsize, 0), (0, 1)),
                ((7, 1, -3), (sys.maxsize, 0, -2), (0, 1, -1)),
            ),
        ),
    )
    def it_computes_the_insertion_orderings_to_help(
        self, request, insertion_positions, expected_value
    ):
        subtotals_ = tuple(
            instance_mock(request, _Subtotal) for _ in insertion_positions
        )
        property_mock(
            request, _BaseAnchoredCollator, "_subtotals", return_value=subtotals_
        )
        _insertion_position_ = method_mock(
            request,
            _BaseAnchoredCollator,
            "_insertion_position",
            side_effect=iter(insertion_positions),
        )
        collator = _BaseAnchoredCollator(None, None)

        insertion_orderings = collator._insertion_orderings

        assert _insertion_position_.call_args_list == [
            call(collator, s) for s in subtotals_
        ]
        assert insertion_orderings == expected_value

    @pytest.mark.parametrize(
        "anchor, expected_value",
        (
            ("top", (-1, 0)),
            ("bottom", (sys.maxsize, 0)),
            (666, (sys.maxsize, 0)),
            (1, (0, 1)),
            (2, (1, 1)),
            (3, (2, 1)),
        ),
    )
    def it_can_compute_the_insertion_position_for_a_subtotal_to_help(
        self, request, anchor, expected_value
    ):
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_element_positions_by_id",
            return_value={1: 0, 2: 1, 3: 2},
        )
        subtotal_ = instance_mock(request, _Subtotal, anchor=anchor)
        collator = _BaseAnchoredCollator(None, None)

        assert collator._insertion_position(subtotal_) == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class DescribeExplicitOrderCollator:
    """Unit-test suite for `cr.cube.collator.ExplicitOrderCollator` object."""

    @pytest.mark.parametrize(
        "ids, are_derived, element_positions, expected_value",
        (
            # --- 0 derived elements
            (
                ("s1", "s2"),
                (False, False),
                (),
                (),
            ),
            # --- 2 derived elements, 1 non-derived
            (
                ("s1", "s2", "s3"),
                (True, False, True),
                ((2, 0), (-1, 0)),
                ((2, 0, 0), (-1, 0, 2)),
            ),
        ),
    )
    def it_computes_the_derived_element_orderings_to_help(
        self, request, ids, are_derived, element_positions, expected_value
    ):
        derived_elements_ = tuple(
            instance_mock(request, _Element, element_id=id_, derived=is_derived)
            for id_, is_derived in zip(ids, are_derived)
        )
        property_mock(
            request, _BaseAnchoredCollator, "_elements", return_value=derived_elements_
        )
        _derived_element_position_ = method_mock(
            request,
            ExplicitOrderCollator,
            "_derived_element_position",
            side_effect=iter(element_positions),
        )
        collator = ExplicitOrderCollator(None, None)

        derived_orderings = collator._derived_element_orderings

        assert _derived_element_position_.call_args_list == [
            call(collator, s.element_id) for s in derived_elements_ if s.derived
        ]
        assert derived_orderings == expected_value

    @pytest.mark.parametrize(
        "element_id, anchor, expected_value",
        (
            (1, "top", (-1, 0)),
            (2, "bottom", (sys.maxsize, 0)),
            (3, {"alias": "a", "position": "before"}, (5, -1)),
            (4, {"alias": "c", "position": "after"}, (2, 1)),
            (2, None, (sys.maxsize, 0)),
        ),
    )
    def it_computes_the_derived_element_position_to_help(
        self, request, element_id, anchor, expected_value
    ):
        element_ = instance_mock(request, _Element, anchor=anchor)
        get_by_id_ = method_mock(
            request, _ValidElements, "get_by_id", return_value=element_
        )
        valid_elements_ = _ValidElements(None, None)
        property_mock(
            request, ExplicitOrderCollator, "_elements", return_value=valid_elements_
        )
        property_mock(
            request,
            _BaseAnchoredCollator,
            "_element_positions_by_id",
            return_value={"a": 5, "b": 1, "c": 2},
        )
        collator = ExplicitOrderCollator(None, None)

        actual_value = collator._derived_element_position(element_id)

        assert actual_value == expected_value
        get_by_id_.assert_called_once_with(valid_elements_, element_id)

    @pytest.mark.parametrize(
        "element_ids, element_id_order, expected_value",
        (
            ((), [], ()),
            (
                (1, 2, 3),
                [2, 3, 1],
                ((0, 1, 2), (1, 2, 3), (2, 0, 1)),
            ),
            # --- element not mentioned in transform appears at end in payload order ---
            ((1, 2, 3), [], ((0, 0, 1), (1, 1, 2), (2, 2, 3))),
            ((1, 2, 3), [2], ((0, 1, 2), (1, 0, 1), (2, 2, 3))),
            # --- id appearing in transform but not dimension is ignored ---
            ((), [3, 2, 1], ()),
            (
                (1, 2, 3),
                [2, 4, 1],
                ((0, 1, 2), (1, 0, 1), (2, 2, 3)),
            ),
            # --- id repeated in transform is only used on first encounter ---
            (
                (1, 2, 3),
                [2, 2, 2],
                ((0, 1, 2), (1, 0, 1), (2, 2, 3)),
            ),
            (
                (1, 2, 3),
                [3, 1, 3],
                ((0, 2, 3), (1, 0, 1), (2, 1, 2)),
            ),
        ),
    )
    def it_computes_the_element_order_descriptors_to_help(
        self, request, element_ids, element_id_order, expected_value
    ):
        valid_elements_ = [
            instance_mock(request, _Element, element_id=id, derived=False)
            for id in element_ids
        ]
        dimension_ = instance_mock(request, Dimension, valid_elements=valid_elements_)
        order_spec_ = instance_mock(request, _OrderSpec, element_ids=element_id_order)
        property_mock(
            request,
            ExplicitOrderCollator,
            "_order_spec",
            return_value=order_spec_,
        )
        collator = ExplicitOrderCollator(dimension_, None)

        assert collator._element_order_descriptors == expected_value


class DescribePayloadOrderCollator:
    """Unit-test suite for `cr.cube.collator.PayloadOrderCollator` object."""

    @pytest.mark.parametrize(
        "element_ids, expected_value",
        (
            # --- no (valid) elements in dimension (not an expected case) ---
            ((), ()),
            # --- 1 element ---
            ((302,), ((0, 0, 302),)),
            # --- 3 elements in dimension ---
            ((47, 103, 18), ((0, 0, 47), (1, 1, 103), (2, 2, 18))),
        ),
    )
    def it_computes_the_element_order_descriptors_to_help(
        self, request, element_ids, expected_value
    ):
        property_mock(
            request, PayloadOrderCollator, "_element_ids", return_value=element_ids
        )
        collator = PayloadOrderCollator(None, None)

        assert collator._element_order_descriptors == expected_value


class DescribeSortByValueCollator:
    """Unit-test suite for `cr.cube.collator.SortByValueCollator` object.

    SortByValueCollator computes element ordering for sort-by-value order transforms. It
    is used for all sort-by-value cases, performing the sort based on a base-vector
    provided to it on construction. This base vector can be a "body" vector, a subtotal,
    or a marginal.
    """

    def it_provides_an_interface_classmethod(self, request):
        dimension_ = instance_mock(request, Dimension)
        _init_ = initializer_mock(request, SortByValueCollator)
        property_mock(
            request,
            SortByValueCollator,
            "_display_order",
            return_value=(-3, -1, -2, 1, 0, 2, 3),
        )
        element_values = [2, 3, 1, 0]
        subtotal_values = [9, 7, 8]
        empty_idxs = [3]

        display_order = SortByValueCollator.display_order(
            dimension_, element_values, subtotal_values, empty_idxs
        )

        _init_.assert_called_once_with(
            ANY, dimension_, element_values, subtotal_values, empty_idxs
        )
        assert display_order == (-3, -1, -2, 1, 0, 2, 3)

    def it_computes_the_display_order_to_help(
        self, request, _top_fixed_idxs_prop_, _bottom_fixed_idxs_prop_
    ):
        property_mock(
            request, SortByValueCollator, "_top_subtotal_idxs", return_value=(-2, -3)
        )
        _top_fixed_idxs_prop_.return_value = (3,)
        property_mock(
            request, SortByValueCollator, "_body_idxs", return_value=(2, 1, 5)
        )
        _bottom_fixed_idxs_prop_.return_value = (0, 4)
        property_mock(
            request, SortByValueCollator, "_bottom_subtotal_idxs", return_value=()
        )
        property_mock(request, SortByValueCollator, "_hidden_idxs", return_value=(5,))
        collator = SortByValueCollator(None, None, None, None)

        assert collator._display_order

    @pytest.mark.parametrize(
        "top_fixed_idxs, bottom_fixed_idxs, descending, element_values, expected_value",
        (
            # --- ascending sort ---
            ((), (), False, (8.0, 2.0, 4.0, 1.0), (3, 1, 2, 0)),
            # --- ascending with fixed on top (which therefore do not appear ---
            ((2, 1), (), False, (8.0, 2.0, 4.0, 1.0), (3, 0)),
            # --- descending sort ---
            ((), (), True, (8.0, 2.0, 4.0, 1.0), (0, 2, 1, 3)),
            # --- descending with fixed on bottom ---
            ((), (1, 0), True, (8.0, 2.0, 4.0, 1.0), (2, 3)),
            # --- descending with both kinds of fixed idxs ---
            ((0,), (3,), True, (8.0, 2.0, 4.0, 1.0), (2, 1)),
            # --- ascending sort with nans (at end in order) ---
            ((), (), False, (8.0, np.nan, 2.0, np.nan), (2, 0, 1, 3)),
            # --- descending sort with nans (at end in order) ---
            ((), (), True, (8.0, np.nan, 2.0, np.nan), (0, 2, 1, 3)),
        ),
    )
    def it_computes_the_sorted_body_idxs_to_help(
        self,
        _top_fixed_idxs_prop_,
        _bottom_fixed_idxs_prop_,
        _descending_prop_,
        top_fixed_idxs,
        bottom_fixed_idxs,
        descending,
        element_values,
        expected_value,
    ):
        """Body-idxs are for elements that are not subtotals and not fixed."""
        _top_fixed_idxs_prop_.return_value = top_fixed_idxs
        _bottom_fixed_idxs_prop_.return_value = bottom_fixed_idxs
        _descending_prop_.return_value = descending
        collator = SortByValueCollator(None, element_values, None, None)

        assert collator._body_idxs == expected_value

    def it_computes_the_bottom_exclusion_idxs_to_help(
        self, _iter_fixed_idxs_, _order_spec_prop_, order_spec_
    ):
        _iter_fixed_idxs_.return_value = (n for n in (4, 0, 5, 2))
        _order_spec_prop_.return_value = order_spec_
        order_spec_.bottom_fixed_ids = (0, 1, 2)
        collator = SortByValueCollator(None, None, None, None)

        bottom_exclusion_idxs = collator._bottom_fixed_idxs

        _iter_fixed_idxs_.assert_called_once_with(collator, (0, 1, 2))
        assert bottom_exclusion_idxs == (4, 0, 5, 2)

    @pytest.mark.parametrize(
        "descending, subtotal_idxs, expected_value",
        (
            # --- ascending sort, all subtotals at bottom ---
            (False, (-3, -1, -2), (-3, -1, -2)),
            # --- descending sort, no subtotals at bottom ---
            (True, (-3, -1, -2), ()),
        ),
    )
    def it_computes_the_bottom_subtotal_idxs_to_help(
        self,
        _descending_prop_,
        _subtotal_idxs_prop_,
        descending,
        subtotal_idxs,
        expected_value,
    ):
        _descending_prop_.return_value = descending
        _subtotal_idxs_prop_.return_value = subtotal_idxs
        collator = SortByValueCollator(None, None, None, None)

        assert collator._bottom_subtotal_idxs == expected_value

    @pytest.mark.parametrize(
        "fixed_element_ids, expected_value",
        (((), ()), ((1, 3), (0, 2)), ((1, 3, 7), (0, 2)), ((4, 2), (3, 1))),
    )
    def it_can_iterate_the_fixed_idxs_to_help(
        self, request, fixed_element_ids, expected_value
    ):
        property_mock(
            request, SortByValueCollator, "_element_ids", return_value=(1, 2, 3, 4, 5)
        )
        collator = SortByValueCollator(None, None, None, None)

        assert tuple(collator._iter_fixed_idxs(fixed_element_ids)) == expected_value

    @pytest.mark.parametrize(
        "descending, expected_value",
        (
            (True, True),
            (False, False),
        ),
    )
    def it_knows_whether_the_sort_direction_is_descending_to_help(
        self, _order_spec_prop_, order_spec_, descending, expected_value
    ):
        _order_spec_prop_.return_value = order_spec_
        order_spec_.descending = descending
        collator = SortByValueCollator(None, None, None, None)

        assert collator._descending == expected_value

    @pytest.mark.parametrize(
        "value, expected",
        ((0, False), (1.1, False), ("a", False), (np.nan, True)),
    )
    def it_can_compute_is_nan_to_help(self, value, expected):
        assert SortByValueCollator._is_nan(value) == expected

    @pytest.mark.parametrize(
        "subtotal_values, descending, expected_value",
        (
            ((1, 2, 3), True, (-1, -2, -3)),
            ((1, 2, 3), False, (-3, -2, -1)),
            # --- NaN values fall to end of sequence in payload order ---
            ((1, np.nan, 2, np.nan, 3), True, (-1, -3, -5, -4, -2)),
            # --- regardless of collation-order ---
            ((1, np.nan, 2, np.nan, 3), False, (-5, -3, -1, -4, -2)),
        ),
    )
    def it_computes_the_subtotal_idxs_to_help(
        self, _descending_prop_, descending, subtotal_values, expected_value
    ):
        _descending_prop_.return_value = descending
        collator = SortByValueCollator(None, None, subtotal_values, None)

        assert collator._subtotal_idxs == expected_value

    def it_computes_the_top_fixed_idxs_to_help(
        self, _order_spec_prop_, order_spec_, _iter_fixed_idxs_
    ):
        _order_spec_prop_.return_value = order_spec_
        order_spec_.top_fixed_ids = (4, 2)
        _iter_fixed_idxs_.return_value = (i for i in (1, 2, 4))
        collator = SortByValueCollator(None, None, None, None)

        top_exclusion_idxs = collator._top_fixed_idxs

        _iter_fixed_idxs_.assert_called_once_with(collator, (4, 2))
        assert top_exclusion_idxs == (1, 2, 4)

    @pytest.mark.parametrize(
        "subtotal_idxs, descending, expected_value",
        (((-3, -1, -2), True, (-3, -1, -2)), ((-1, -2, -3), False, ())),
    )
    def it_computes_the_top_subtotal_idxs_to_help(
        self,
        _subtotal_idxs_prop_,
        subtotal_idxs,
        _descending_prop_,
        descending,
        expected_value,
    ):
        _subtotal_idxs_prop_.return_value = subtotal_idxs
        _descending_prop_.return_value = descending
        collator = SortByValueCollator(None, None, None, None)

        assert collator._top_subtotal_idxs == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _bottom_fixed_idxs_prop_(self, request):
        return property_mock(request, SortByValueCollator, "_bottom_fixed_idxs")

    @pytest.fixture
    def _descending_prop_(self, request):
        return property_mock(request, SortByValueCollator, "_descending")

    @pytest.fixture
    def _iter_fixed_idxs_(self, request):
        return method_mock(request, SortByValueCollator, "_iter_fixed_idxs")

    @pytest.fixture
    def order_spec_(self, request):
        return instance_mock(request, _OrderSpec)

    @pytest.fixture
    def _order_spec_prop_(self, request):
        return property_mock(request, SortByValueCollator, "_order_spec")

    @pytest.fixture
    def _subtotal_idxs_prop_(self, request):
        return property_mock(request, SortByValueCollator, "_subtotal_idxs")

    @pytest.fixture
    def _top_fixed_idxs_prop_(self, request):
        return property_mock(request, SortByValueCollator, "_top_fixed_idxs")
